#include "src/core/armips.hpp"

#include <sstream>
#include <regex>
#include <iterator>
#include <cstdio>
#include <QFile>
#include "src/polyfill/appdata.hpp"
#include "src/polyfill/filesystem.hpp"
#include "src/polyfill/filestream.hpp"
#include "src/polyfill/exec.hpp"
#include "src/core/blueprint.hpp"
#include "src/core/util.hpp"
#include "src/core/storage.hpp"
#include "src/core/exceptions.hpp"
#include "src/core/checksum.hpp"
#include "src/core/asm-extensions.hpp"
#include "src/ui/util.hpp"

static const std::regex s_romNameReplacer(
	"\\$\\{ROM_NAME\\}",
	std::regex_constants::ECMAScript | std::regex_constants::optimize
);

static const std::regex s_asmDirReplacer(
	"\\$\\{ASM_DIR\\}",
	std::regex_constants::ECMAScript | std::regex_constants::optimize
);

#ifdef _WIN32
	static const string ARMIPS_EXE = "armips.exe";
#else
	static const string ARMIPS_EXE = "armips";
#endif

bool Armips::exists() {
	const fs::path binPath = AppData::dataDir() / ARMIPS_EXE;
	if( !fs::exists( binPath ) ) {
		return false;
	}

	// The old bundled version had a bug with .loadtable, so it needs to be updated
#ifndef _WIN32
	static const md5sum badVersion = { 0xB7, 0x94, 0xBC, 0x70, 0x3B, 0xA9, 0xC4, 0xB0, 0x0C, 0xAD, 0x11, 0x73, 0x00, 0x29, 0x1E, 0xB3 };
	FileReadStream file( binPath.u8string() );
	if( getChecksumMD5( file ) == badVersion ) {
		file.close();
		fs::forceDelete( binPath );
		return false;
	}
#endif

	return true;
}

static const char *const s_skeletons[] = {
	"defines.asm",
	"on-game-start.asm",
	"on-mario-step-begin.asm",
	"on-mario-step-end.asm",
	"on-frame-end.asm",
	"on-hud-render.asm",
	"on-level-load.asm",
	"on-area-load.asm",
	"misc-injected.asm",
	"behaviour-script-linkers.asm",
	"global.asm"
};
static constexpr size_t s_numSkeletons = sizeof( s_skeletons ) / sizeof( const char* );

static const char s_referencesHeader[] = ""
	"/*\n"
	"This file was autogenerated by Bowser's Blueprint. Do not modify the file, as\n"
	"contents will be overwritten next time you build.\n"
	"\n"
	"This file contains labels referencing data generated by Bowser's Blueprints.\n"
	"*/";

static const char s_triggersHeader[] = ""
	"/*\n"
	"This file was autogenerated by Bowser's Blueprint. Do not modify the file, as\n"
	"contents will be overwritten next time you build.\n*/\n";

void Armips::autogenerateWithoutReferences() {
	const string romName = Blueprint::current()->targetRomName();

	const fs::path asmDir = Blueprint::current()->getAsmDirectory();
	const fs::path asmFilePath = Blueprint::current()->getAsmFilePath();
	fs::create_directory( asmDir );

	QFile masterInput( ":/asm/master.asm" );
	masterInput.open( QIODevice::ReadOnly );

	const string masterAsm = std::regex_replace(
		std::regex_replace(
			QString( masterInput.readAll() ).toStdString(),
			s_romNameReplacer,
			romName
		),
		s_asmDirReplacer,
		asmDir.stem().u8string()
	);
	FileWriteStream masterOutput( asmFilePath.u8string(), std::ios_base::out | std::ios_base::trunc );
	masterOutput.write( masterAsm.c_str(), masterAsm.length() );

	for( size_t i = 0; i < s_numSkeletons; i++ ) {
		const fs::path filePath = asmDir / s_skeletons[i];
		if( !fs::exists( filePath ) ) {
			QResourceStream input( (string(":/asm/") + s_skeletons[i]).c_str() );
			FileWriteStream output( filePath.u8string(), std::ios_base::out | std::ios_base::trunc );
			while( input.peek() != std::char_traits<char>::eof() ) output.put( input.get() );
		}
	}

	if( !fs::exists( asmDir / "bb-refs.asm" ) ) {
		FileWriteStream refsFile( (asmDir / "bb-refs.asm").u8string(), std::ios_base::out );
		refsFile << s_referencesHeader << std::endl << std::flush;
	}

	if( !fs::exists( asmDir / "camera-triggers.asm" ) ) {
		FileWriteStream refsFile( (asmDir / "camera-triggers.asm").u8string(), std::ios_base::out );
		refsFile << s_triggersHeader << std::endl << std::flush;
	}

	clearReadOnlyFlagOnWindows( asmDir );
}

void Armips::autogenerate(
	AsmReferences asmReferences,
	const std::map<string,uint> &cameraTriggerRefs
) {
	const fs::path asmDir = Blueprint::current()->getAsmDirectory();

	autogenerateWithoutReferences();
	FileWriteStream refsFile( (asmDir / "bb-refs.asm").u8string(), std::ios_base::out | std::ios_base::trunc );
	refsFile << s_referencesHeader << std::endl << std::endl;
	for( const auto &i : asmReferences ) {
		if( i.second.empty() ) continue;
		refsFile << "; Level " << (uint)i.first << " (" + Blueprint::current()->levels().at( i.first ).name << ")" << std::endl;
		for( const auto &j : i.second ) {
			char hexAddress[11];
			std::sprintf( hexAddress, "0x%X", j.second );
			refsFile << "bb_level_" << (int)i.first << "_" << j.first << " equ " << hexAddress << std::endl;
		}
		refsFile << std::endl;
	}
	refsFile << std::flush;

	FileWriteStream triggersFile( (asmDir / "camera-triggers.asm").u8string(), std::ios_base::out | std::ios_base::trunc );
	triggersFile << s_triggersHeader << std::endl << std::endl;

	triggersFile << ".orga " << (uint)AsmExtensions::customCameraTablePointerRomLocation << std::endl;
	triggersFile << "LUI T0, hi( bbp_custom_camera_init_table - 0x48 )" << std::endl;
	triggersFile << "ORI T0, T0, lo( bbp_custom_camera_init_table - 0x48 )" << std::endl;

	for( const auto &i : cameraTriggerRefs ) {
		char orgaCmd[17];
		std::sprintf( orgaCmd, ".orga 0x%X", i.second );
		triggersFile << orgaCmd << std::endl;
		triggersFile << ".word " << i.first << std::endl << std::endl;
	}
	triggersFile << std::flush;
}

void Armips::build() {
	fs::create_directories( AppData::dataDir() );
	const string armipsCmd = quoteAndEscape( (AppData::dataDir() / ARMIPS_EXE).u8string() );

	string workingDirectory;
	if( Blueprint::current()->isCollab() ) {
		workingDirectory = Blueprint::current()->getAsmDirectory().parent_path().u8string();
	} else {
		workingDirectory = Blueprint::current()->targetRomPath().parent_path().u8string();
	}

	const string cmd = armipsCmd + " " + quoteAndEscape( Blueprint::current()->getAsmFilePath().u8string() );

	std::error_code err;
	string armipsOutput;
	if( !Command::execSync( cmd, workingDirectory, armipsOutput, err ) ) {
		throw ArmipsException( armipsOutput );
	}
}

static const std::regex s_hexAddr_regex(
	"^0x[0-9A-F]{1,8}$",
	std::regex_constants::ECMAScript | std::regex_constants::optimize
);

static const std::regex s_label_regex(
	"^bb_level_(\\d{1,2})_(\\S+)$",
	std::regex_constants::ECMAScript | std::regex_constants::optimize
);

AsmReferences Armips::getCurrentAsmReferences() {
	const fs::path asmRefFilePath = Blueprint::current()->getAsmDirectory() / "bb-refs.asm";

	AsmReferences references;
	if( !fs::exists( asmRefFilePath ) ) {
		return references;
	}

	FileReadStream fileStream( asmRefFilePath.u8string(), std::ios_base::in );
	while( fileStream.good() && !fileStream.eof() ) {
		string lineStr;
		std::getline( fileStream, lineStr );
		if( lineStr.empty() ) {
			continue;
		}

		string label, equ, addr;
		std::istringstream line( lineStr );
		line >> label >> equ >> addr;

		std::smatch matches;
		if(
			equ != "equ" ||
			!std::regex_match( addr, s_hexAddr_regex ) ||
			!std::regex_match( addr, matches, s_label_regex ) ||
			matches.size() < 3
		) continue;

		const LevelId level = (LevelId)std::stoi( matches[1] );
		const string ref = matches[2];
		const uint value = (uint)std::stol( addr.substr( 2 ), nullptr, 16 );

		references[level][ref] = value;
	}

	return references;
}
