#include "src/core/asm-modules.hpp"

#include <cassert>
#include <cstdlib>
#include <cstdio>
#include <cctype>
#include <locale>
#include <unordered_set>
#include <unordered_map>
#include <set>
#include <QFile>
#include "src/core/json.hpp"
#include "src/core/exceptions.hpp"
#include "src/core/baserom.hpp"
#include "src/core/util.hpp"
#include "src/core/package.hpp"
#include "src/openmode.hpp"
#include "src/polyfill/filestream.hpp"
#include "src/polyfill/appdata.hpp"

#ifdef _WIN32
#include <strstream>
#include <sstream>
#include "src/polyfill/buffer.hpp"
#endif

#include "src/core/asm-modules-json.cpp.inc"

enum class AsmParserState {
	Normal,
	MaybeStartComment,
	BlockComment,
	LineComment,
	Escape,
	String,
	MaybeEndComment
};

static const char *s_header = ""
"/* This file was autogenerated by Bowser's Blueprints. Do not modify this file,\n"
" * as any changes will be lost upon rebuilding the blueprint. */\n";

static const string SHIM_END = "0_END";

static void processAsmFile(
	const fs::path &sourcePath,
	const fs::path &outputPath,
	const Uuid &moduleId,
	const HashMap<string,string> &nextShim
) {
#ifdef _WIN32
	const size_t sourceFileSize = fs::file_size( sourcePath );
	FileReadStream sourceFile( sourcePath.u8string() );
	Buffer buffer( sourceFileSize + 1 );
	sourceFile.read( buffer.data(), sourceFileSize );
	buffer.data()[sourceFileSize] = '\0';
	std::istrstream source( buffer.data() );
	std::ostringstream output;

#else
	FileReadStream source( sourcePath.u8string() );
	FileWriteStream output( outputPath.u8string() );
#endif

	AsmParserState state = AsmParserState::Normal;
	char c = source.get();
	while( source.good() && output.good() && !source.eof() ) {
		switch( state ) {
			case AsmParserState::Normal: {
				switch( c ) {
					case '$': {
						output << "bbp_" << moduleId.toHex() << "_";
						c = source.get();
						continue;
					}
					case '#': {
						string cmd;
						source >> cmd;
						Util::makeLowerCase( cmd );

						if( cmd == "continue" ) {
							string arg;
							source >> arg;
							arg = trimStart( arg );

							const string lookup = "bbp_"s + moduleId.toHex() + "_" + arg;
							if( !nextShim.count( lookup ) ) {
								throw AsmModuleException( "Invalid argument to #continue: shim '"s + arg + "' not found." );
							}

							const string &next = nextShim.at( lookup );
							if( next == SHIM_END ) {
								output << "JR RA";
							} else {
								output << "J " << next;
							}
						} else if( cmd == "await" ) {
							string arg;
							source >> arg;
							arg = trimStart( arg );

							const string lookup = "bbp_"s + moduleId.toHex() + "_" + arg;
							if( !nextShim.count( lookup ) ) {
								throw AsmModuleException( "Invalid argument to #await: shim '"s + arg + "' not found." );
							}

							const string &next = nextShim.at( lookup );
							if( next == SHIM_END ) {
								output << "NOP";
							} else {
								output << "JAL " << next;
							}

						} else throw AsmModuleException( "Unknown command #"s + cmd );

						c = source.get();
						continue;
					}
					case ';': state = AsmParserState::LineComment; break;
					case '/': state = AsmParserState::MaybeStartComment; break;
					case '"': state = AsmParserState::String; break;
					default: break;
				}
				break;
			}
			case AsmParserState::MaybeStartComment: {
				switch( c ) {
					case '/': state = AsmParserState::LineComment; break;
					case '*': state = AsmParserState::BlockComment; break;
					default: state = AsmParserState::Normal; continue;
				}
				break;
			}
			case AsmParserState::BlockComment: {
				if( c == '*' ) {
					state = AsmParserState::MaybeEndComment;
				}
				break;
			}
			case AsmParserState::LineComment: {
				if( c == '\n' ) {
					state = AsmParserState::Normal;
				}
				break;
			}
			case AsmParserState::Escape: {
				state = AsmParserState::String;
				break;
			}
			case AsmParserState::String: {
				if( c == '"' ) {
					state = AsmParserState::Normal;
				} else if( c == '\\' ) {
					state = AsmParserState::Escape;
				}
				break;
			}
			case AsmParserState::MaybeEndComment: {
				state = (c == '/') ? AsmParserState::Normal : AsmParserState::BlockComment;
				break;
			}
			default: assert( false );
		}

		output.put( c );
		c = source.get();
	}

	if( source.bad() && !source.eof() ) {
		throw AsmModuleException( "Failed to open asm file for reading" );
	} else if( output.bad() ) {
		throw AsmModuleException( "Failed to open or create asm file for writing" );
	}

#ifdef _WIN32
	FileWriteStream outputFile( outputPath.u8string() );
	outputFile.write( output.str().data(), output.str().size() );
#endif

}

static void processFolder(
	const fs::path &sourceDir,
	const fs::path &outputDir,
	const Uuid &moduleId,
	const HashMap<string,string> &shimPaths
) {
	fs::create_directories( outputDir );
	for( const fs::directory_entry &entry : fs::directory_iterator( sourceDir ) ) {
		if( entry.status().type() == fs::file_type::directory ) {
			const string dirName = entry.path().filename().u8string();
			processFolder( sourceDir / dirName, outputDir / dirName, moduleId, shimPaths );
			continue;
		}

		if(
			entry.status().type() != fs::file_type::regular ||
			entry.path().filename() == "module.json"
		) continue;

		const fs::path outputFile = outputDir / entry.path().filename();
		if( entry.path().extension().u8string() == ".asm" ) {
			processAsmFile( entry.path(), outputFile, moduleId, shimPaths );
		} else {
			fs::tryCopyFileOverwrite( entry.path(), outputFile );
		}

	}
	clearReadOnlyFlagOnWindows( outputDir );
}

static inline string getShimTarget( ShimType type, uint romAddr ) {
	uint word = BaseRom::getWord( romAddr );
	switch( type ) {
		case ShimType::Word: {
			if( word < 0x80000000 || word >= 0x80800000 || word % 4 != 0 ) {
				throw AsmModuleException( "Invalid Word shim: original value is not a valid function pointer." );
			}
			return Util::toHexWord( word, true );
		}
		case ShimType::Head: {
			if( word >> 26 != 3 ) {
				throw AsmModuleException( "Invalid Head shim: instruction is not a JAL instruction." );
			}
			return Util::toHexWord( 0x80000000 | ((word & 0x3FFFFFF) << 2), true );
		}
		case ShimType::Tail: {
			if( word != 0x03E00008 ) {
				throw AsmModuleException( "Invalid Tail shim: instruction is not JR RA instruction." );
			}
			return SHIM_END;
		}
		case ShimType::Jump: {
			if( word >> 26 != 2 ) {
				throw AsmModuleException( "Invalid Head shim: instruction is not a J instruction." );
			}
			return Util::toHexWord( 0x80000000 | ((word & 0x3FFFFFF) << 2), true );
		}
		default: throw AsmModuleException( "Unknown shim type" );
	}
}

struct CustomActionInfo {
	string varName;
	string codeLabel;
};

struct CustomActionFileInfo {
	string name;
	string breakAddr;
	string continueAddr;
	string restoreReg;
	string restoreAT;
};

static const CustomActionFileInfo s_actionInfo[7] = {
	{ "mario_actions_stationary", "0x80263E08", "0x802638E4", "T6", "0x800" },
	{ "mario_actions_moving", "0x80269ED0", "0x802699A0", "T6", "0x8C" },
	{ "mario_actions_airborne", "0x802700E0", "0x8026FB3C", "T6", "0x180" },
	{ "mario_actions_submerged", "0x80274E88", "0x802749F0", "T1", "0x3000" },
	{ "mario_actions_cutscene", "0x8025DD14", "0x8025D7C8", "T6", "0x2" },
	{ "mario_actions_automatic", "0x80260888", "0x8026060C", "T7", "0x10" },
	{ "mario_actions_object", "0x8027617C", "0x8027602C", "T6", "0x80" }
};

static void createCustomActionFile(
	const fs::path &moduleDir,
	MarioActionCategory category,
	const std::vector<CustomActionInfo> actions
) {
	const uint actionIndex = (uint)category / 0x40;
	assert( actionIndex < 7 );
	const CustomActionFileInfo &fileInfo = s_actionInfo[actionIndex];

	FileWriteStream file( (moduleDir / (fileInfo.name + ".asm")).u8string() );
	file << s_header << std::endl;

	file << "bbp_" << fileInfo.name << ":" << std::endl;
	if( actions.empty() ) {
		file << "LW " << fileInfo.restoreReg << ", 0x28 (SP)" << std::endl;
		file << "J " << fileInfo.continueAddr << std::endl;
		file << "LUI AT, " << fileInfo.restoreAT << std::endl;
		return;
	}

	file << "LI RA, @@break" << std::endl;
	file << "LW T0, 0x8033B17C" << std::endl;

	int i = 0;
	for( const CustomActionInfo &action : actions ) {
		file << "LI T1, " << action.varName << std::endl;
		file << "BNE T0, T1, @@endif_" << ++i << std::endl;
		file << "NOP" << std::endl;
		file << "J " << action.codeLabel << std::endl;
		file << "LW A0, 0x28 (SP)" << std::endl;
		file << "@@endif_" << i << ":" << std::endl;
	}

	file << "LW " << fileInfo.restoreReg << ", 0x28 (SP)" << std::endl;
	file << "J " << fileInfo.continueAddr << std::endl;
	file << "LUI AT, " << fileInfo.restoreAT << std::endl;

	file << "@@break:" << std::endl;
	file << "J " << fileInfo.breakAddr << std::endl;
	file << "SW V0, 0x24 (SP)" << std::endl;
}

template<typename T> static inline const T &getOrDefault(
	const AsmModule &module,
	const AsmModuleProperty &property
) {
	if( const AsmModulePropertyValue *const value = Util::tryGetValue( module.parameters, property.name ) ) {
		if( std::holds_alternative<T>( *value ) ) {
			return std::get<T>( *value );
		}
	}
	return std::get<T>( property.defaultValue );
}

inline static string escapeNewlines( const string &str ) {
	string out = "";
	for( char c : str ) {
		switch( c ) {
			case '\n': out += "\\n"; break;
			case '\r': break;
			default: out += c;
		}
	}
	return out;
}

struct SortableHook {
	string label;
	HookPriority priority;

	bool operator<( const SortableHook &other ) const {
		if( priority != other.priority ) {
			return priority < other.priority;
		} else {
			return label < other.label;
		}
	}
};

struct ShimChain {
	ShimType type;
	sbyte minNiceness;
	std::set<SortableHook> hooks;
};

static const std::pair<ModuleHook,string> s_hookFiles[7] = {
	{ ModuleHook::MarioStepBegin, "on-mario-step-begin.asm" },
	{ ModuleHook::MarioStepEnd, "on-mario-step-end.asm" },
	{ ModuleHook::LevelLoad, "on-level-load.asm" },
	{ ModuleHook::AreaLoad, "on-area-load.asm" },
	{ ModuleHook::GameStart, "on-game-start.asm" },
	{ ModuleHook::RenderHUD, "on-hud-render.asm" },
	{ ModuleHook::FrameEnd, "on-frame-end.asm" }
};

static inline string prefixLabel( const Uuid &moduleId, const string &label ) {
	return "bbp_"s + moduleId.toHex() + "_" + label;
}

static inline string sanitizeCameraName( const string cameraName ) {
	string cameraLabel = "bbp_camera_id_";
	for( char c : cameraName ) {
		if( c == ' ' || c == '_' ) {
			cameraLabel += '_';
		} else if(
			(c >= 'a' && c <= 'z') ||
			(c >= 'A' && c <= 'Z') ||
			(c >= '0' && c <= '9')
		) {
			cameraLabel += c;
		}
	}
	return cameraLabel;
}

static const char *s_cameraShim = ""
"bbp_custom_camera_shim:\n"
"ADDIU T6, R0, 0x400\n"
"LUI AT, 0x8033\n"
"SH T6, 0xDF34 (AT)\n"
"LW A0, 0x40 (SP)\n"
"LB T6, 0x0 (A0)\n"
"ADDIU T6, T6, 0xFFEE\n"
"SLT AT, T6, R0\n"
"BNE AT, R0, @@vanilla_cam\n"
"SLL T6, T6, 2\n"
"LI T5, @camera_loop_table\n"
"ADDU T6, T5, T6\n"
"LW T6, 0x0 (T6)\n"
"ADDIU SP, SP, 0xFFE8\n"
"SW RA, 0x14 (SP)\n"
"JALR T6\n"
"NOP\n"
"LW RA, 0x14 (SP)\n"
"J 0x80286D64\n"
"ADDIU SP, SP, 0x18\n"
"@@vanilla_cam:\n"
"J 0x80286BB0\n"
"NOP";

void AsmModuleBuilder::generate(
	const std::map<Uuid,AsmModule> &modules,
	const std::map<string,CustomMarioAction> &actions,
	const std::map<Uuid,CustomCamera> &cameraModules,
	const fs::path &sourceDir,
	const fs::path &outputDir
) {
	fs::forceDeleteRecursive( outputDir );
	fs::create_directories( outputDir );

	std::map<uint,ShimChain> customHooks;
	std::array<std::set<SortableHook>,7> standardHooks;
	std::array<std::vector<CustomActionInfo>,7> customActions;

	/* file scope */ {
		FileWriteStream propsFile( (outputDir / "props.asm").u8string() );
		FileWriteStream defsFile( (outputDir / "definitions.asm").u8string() );
		FileWriteStream stringsFile( (outputDir / "strings.asm").u8string() );
		FileWriteStream injectFile( (outputDir / "injected.asm").u8string() );
		FileWriteStream globalsFile( (outputDir / "global.asm").u8string() );
		FileWriteStream scopesFile( (outputDir / "scopes.asm").u8string() );
		FileWriteStream camerasFile( (outputDir / "cameras.asm").u8string() );

		propsFile.imbue( std::locale::classic() );

		propsFile << s_header << std::endl;
		defsFile << s_header << std::endl;
		stringsFile << s_header << std::endl;
		injectFile << s_header << std::endl;
		globalsFile << s_header << std::endl;
		scopesFile << s_header << std::endl;
		camerasFile << s_header << std::endl;

		stringsFile << ".loadtable \"encoding.tbl\"" << std::endl;
		fs::forceDelete( outputDir / "encoding.tbl" );
		QFile::copy( ":/asm/encoding.tbl", (outputDir / "encoding.tbl").u8string().c_str() );
		clearReadOnlyFlagOnWindows( outputDir / "encoding.tbl" );

		std::array<std::vector<uint>,7> availableActionIds = {{
			{ 0x37, 0x2E, 0x2D, 0x2C, 0x2B, 0x2A, 0x29, 0x28, 0x1F, 0x1E, 0x1D, 0x1C, 0x1B, 0x1A, 0x19, 0x18, 0x17, 0x16, 0x15, 0x14, 0x13, 0x12, 0x11, 0x10, 0xF, 0xE, 0xC },
			{ 0x7F, 0x7E, 0x7D, 0x7C, 0x7B, 0x6F, 0x6E, 0x6D, 0x6C, 0x6B, 0x6A, 0x69, 0x68, 0x5F, 0x5E, 0x5D, 0x5C, 0x5B, 0x58, 0x4F, 0x4E, 0x4D, 0x4C, 0x41 },
			{ 0xBF, 0xBC, 0xBB, 0xBA, 0xB9, 0xA5, 0x9F, 0x9E, 0x9D, 0x97, 0x96, 0x95, 0x93, 0x92, 0x91, 0x90, 0x8F, 0x8B, 0x84 },
			{ 0xFF, 0xFE, 0xFD, 0xFC, 0xEF, 0xEE, 0xED, 0xEC, 0xEB, 0xEA, 0xE9, 0xE8, 0xE7, 0xE6, 0xE5, 0xE4, 0xDF, 0xDE, 0xDD, 0xDC, 0xDB, 0xDA, 0xD9, 0xD8, 0xD7, 0xCF, 0xCE, 0xCD, 0xCC, 0xCB, 0xCA, 0xC9 },
			{ 0x13F, 0x13E, 0x130, 0x11F, 0x11E, 0x11D, 0x11C, 0x11B, 0x110, 0x10E, 0x10D, 0x10C, 0x10B },
			{ 0x17F, 0x17E, 0x17D, 0x17C, 0x17B, 0x17A, 0x179, 0x178, 0x177, 0x176, 0x175, 0x174, 0x173, 0x16F, 0x16E, 0x16D, 0x16C, 0x16B, 0x16A, 0x169, 0x168, 0x167, 0x166, 0x165, 0x164, 0x163, 0x162, 0x161, 0x160, 0x15F, 0x15E, 0x15D, 0x15C, 0x15B, 0x15A, 0x159, 0x158, 0x157, 0x156, 0x155, 0x154, 0x153, 0x152, 0x151, 0x150, 0x147, 0x146 },
			{ 0x1BF, 0x1BE, 0x1BD, 0x1BC, 0x1BB, 0x1BA, 0x1B9, 0x1B8, 0x1B7, 0x1B6, 0x1B5, 0x1B4, 0x1B3, 0x1B2, 0x1B1, 0x1B0, 0x1AF, 0x1AE, 0x1AD, 0x1AC, 0x1AB, 0x1AA, 0x1A9, 0x1A8, 0x1A7, 0x1A6, 0x1A5, 0x1A4, 0x1A3, 0x1A2, 0x1A1, 0x1A0, 0x19F, 0x19E, 0x19D, 0x19C, 0x19B, 0x19A, 0x199, 0x198, 0x197, 0x196, 0x195, 0x194, 0x193, 0x18F, 0x18E, 0x18D, 0x18C, 0x18B, 0x18A, 0x184, 0x182, 0x181 }
		}};

		for( const auto &a : actions ) {
			const CustomMarioAction &action = a.second;
			std::vector<uint> &freeIds = availableActionIds.at( (uint)action.category / 0x40 );
			if( freeIds.empty() ) {
				throw AsmModuleException( "Out of action IDs for "s + Enum::toString<MarioActionCategory>( action.category ) + " group." );
			}

			customActions[(uint)action.category / 0x40].push_back({
				action.variableName,
				action.functionLabel
			});

			const uint actionId = (uint)action.flags | (uint)action.category | freeIds.back();
			freeIds.pop_back();

			defsFile << action.variableName << " equ " << Util::toHexWord( actionId, true ) << std::endl;
		}

		for( const auto &i : modules ) {
			const string idString = i.first.toString();

			const AsmModule &module = i.second;
			const string prefix = "bbp_"s + i.first.toHex() + "_";
			for( const AsmModuleProperty &prop : module.definition.properties ) {
				if( prop.type == ModulePropertyType::String ) {
					stringsFile << prefix << prop.label << ":" << std::endl;
					stringsFile << (prop.encoded ? ".string " : ".asciiz " ) << escapeNewlines( quoteAndEscape( getOrDefault<string>( module, prop ) ) ) << std::endl;
					continue;
				}

				if( prop.type == ModulePropertyType::Colour && prop.colourFormat == ModuleColourFormat::Components ) {
					const ColourRGBA32 &colour = getOrDefault<ColourRGBA32>( module, prop );
					propsFile << prefix << prop.label << "_red equ " << (int)colour.red << std::endl;
					propsFile << prefix << prop.label << "_green equ " << (int)colour.green << std::endl;
					propsFile << prefix << prop.label << "_blue equ " << (int)colour.blue << std::endl;
					propsFile << prefix << prop.label << "_alpha equ " << (int)colour.alpha << std::endl;
					continue;
				}

				propsFile << prefix << prop.label << " equ ";
				switch( prop.type ) {
					case ModulePropertyType::Integer:
						propsFile << getOrDefault<int>( module, prop );
						break;
					case ModulePropertyType::Float:
						propsFile << getOrDefault<float>( module, prop );
						break;
					case ModulePropertyType::Angle:
						propsFile << (ushort)((uint)( (double)getOrDefault<float>( module, prop ) * (double)0x10000 / 360.0 ) % 0x10000);
						break;
					case ModulePropertyType::Boolean:
						propsFile << (getOrDefault<bool>( module, prop ) ? prop.trueValue : prop.falseValue);
						break;
					case ModulePropertyType::Colour: {
						const ColourRGBA32 &colour = getOrDefault<ColourRGBA32>( module, prop );
						if( prop.colourFormat == ModuleColourFormat::RGBA32 ) {
							propsFile << colour.toInt();
						} else {
							assert( prop.colourFormat == ModuleColourFormat::RGBA16 );
							propsFile << ((ColourRGBA5551)colour).toShort();
						}
						break;
					}
					case ModulePropertyType::Pointer: {
						char hexAddr[11];
						std::sprintf( hexAddr, "0x%08X", getOrDefault<uint>( module, prop ) );
						propsFile << hexAddr;
						break;
					}
					default: assert( false );
				}
				propsFile << std::endl;
			}

			if( module.definition.areaScoped ) {
				scopesFile << prefix << "area_mask:" << std::endl;
				scopesFile << ".byte ";
				for( size_t j = 0; j < 39; j++ ) {
					scopesFile << (uint)module.areaMask[j] << ",";
				}
				scopesFile << (uint)module.areaMask[39] << std::endl;
			}

			HashMap<string,MarioActionCategory> actionCategoryMap;
			for( const CustomMarioAction &action : module.definition.actions ) {
				std::vector<uint> &freeIds = availableActionIds.at( (uint)action.category / 0x40 );
				if( freeIds.empty() ) {
					throw AsmModuleException( "Out of action IDs for "s + Enum::toString<MarioActionCategory>( action.category ) + " group." );
				}
				const uint actionId = freeIds.back() | (uint)action.flags;
				freeIds.pop_back();

				defsFile << prefix << action.variableName << " equ " << Util::toHexWord( actionId, true ) << std::endl;
				if( action.isPublic ) {
					defsFile << action.variableName << " equ " << prefix << action.variableName << std::endl;
				}
				actionCategoryMap[action.variableName] = action.category;
			}

			for( const string &file : module.definition.definitionFiles ) {
				defsFile << ".include " << quoteAndEscape( idString + '/' + file ) << std::endl;
			}

			for( const string &file : module.definition.injectedFiles ) {
				injectFile << ".include " << quoteAndEscape( idString + '/' + file ) << std::endl;
			}

			for( const AbsoluteAsmFile &file : module.definition.globalFiles ) {
				assert( file.romEnd >= file.romStart );
				globalsFile << ".orga " << file.romStart << std::endl;
				globalsFile << ".area " << (file.romEnd - file.romStart) << std::endl;
				globalsFile << ".include " << quoteAndEscape( idString + '/' + file.file ) << std::endl;
				globalsFile << ".endarea" << std::endl;
			}

			for( const AsmModuleHook &hook : module.definition.hooks ) {
				if( hook.type == ModuleHook::MarioAction ) {
					const MarioActionCategory group = actionCategoryMap.at( hook.actionRef );
					customActions.at( (uint)group / 0x40 ).push_back({
						prefix + hook.actionRef,
						prefix + hook.label
					});
				} else if( hook.type == ModuleHook::Custom ) {
					ShimChain &shimChain = customHooks[hook.shimAddress];
					if( shimChain.hooks.empty() ) {
						shimChain.type = hook.shimType;
						shimChain.minNiceness = hook.niceness;
					} else {
						if( shimChain.type != hook.shimType ) {
							throw AsmModuleException( "Cannot build ASM Modules while there is a conflict" );
						}

						const sbyte requiredNiceness = 5 - hook.niceness;
						if( shimChain.minNiceness < requiredNiceness ) {
							throw AsmModuleException( "Cannot build ASM Modules while there is a conflict" );
						}

						shimChain.minNiceness = (hook.niceness < shimChain.minNiceness) ? hook.niceness : shimChain.minNiceness;
					}

					shimChain.hooks.insert( SortableHook{
						prefix + hook.label,
						hook.priority
					});
				} else {
					standardHooks.at( (uint)hook.type ).insert( SortableHook{
						prefix + hook.label,
						hook.priority
					});
				}
			}
		}

		stringsFile << ".align 4" << std::endl;

		std::map<short,string> initLabels;
		std::map<short,string> loopLabels;
		for( const auto &i : cameraModules ) {
			initLabels[i.second.cameraId] = prefixLabel( i.second.asmModuleId, i.second.module->initLabel );
			loopLabels[i.second.cameraId] = prefixLabel( i.second.asmModuleId, i.second.module->label );
			defsFile << sanitizeCameraName( i.second.module->name ) << " equ " << (int)i.second.cameraId << std::endl;
		}

		camerasFile << "bbp_custom_camera_init_table:" << std::endl;
		if( initLabels.empty() ) {
			camerasFile << "NOP ; no init functions used" << std::endl;
		} else for( short i = 0x12; i < 0x12 + (short)initLabels.size(); i++ ) {
			camerasFile << ".word ";
			const string &label = initLabels.at( i );
			if( label.empty() ) {
				camerasFile << "0" << std::endl;
			} else {
				camerasFile << label << std::endl;
			}
		}

		camerasFile << std::endl << "@camera_loop_table:" << std::endl;
		if( loopLabels.empty() ) {
			camerasFile << "NOP ; no camera modules used" << std::endl;
		} else for( short i = 0x12; i < 0x12 + (short)loopLabels.size(); i++ ) {
			camerasFile << ".word " << loopLabels.at( i ) << std::endl;
		}

		camerasFile << s_cameraShim << std::flush;

		globalsFile << std::endl << ".orga 0x41BA4" << std::endl;
		globalsFile << "J bbp_custom_camera_shim" << std::endl;
		globalsFile << "NOP :: NOP" << std::endl;
	}

	for( uint i = 0; i < 7; i++ ) {
		createCustomActionFile( outputDir, (MarioActionCategory)(i * 0x40 ), customActions[i] );

		FileWriteStream hookFile( (outputDir / s_hookFiles[i].second).u8string() );
		hookFile << s_header << std::endl;

		const std::set<SortableHook> &hooks = standardHooks.at( (uint)s_hookFiles[i].first );
		for( const SortableHook &hook : hooks ) {
			hookFile << "JAL " << hook.label << std::endl;
			if( s_hookFiles[i].first == ModuleHook::FrameEnd ) {
				hookFile << "LW A0, 0x10 (SP)" << std::endl;
			} else {
				hookFile << "NOP" << std::endl;
			}
		}
	}

	/* Shim scope */ {
		FileWriteStream shimFile( (outputDir / "shims.asm").u8string() );
		shimFile << s_header << std::endl;

		for( const auto &i : customHooks ) {
			shimFile << ".orga " << Util::toHexWord( i.first, true ) << std::endl;
			switch( i.second.type ) {
				case ShimType::Word:
					shimFile << ".word "; break;
				case ShimType::Head:
					shimFile << "JAL "; break;
				case ShimType::Tail:
				case ShimType::Jump:
					shimFile << "J "; break;
				default:
					assert( false );
			}
			shimFile << i.second.hooks.begin()->label << std::endl;
		}
	}

	HashMap<string,string> shimPath;
	for( const auto &i : customHooks ) {
		string next = getShimTarget( i.second.type, i.first );
		for( auto j = i.second.hooks.rbegin(); j != i.second.hooks.rend(); j++ ) {
			shimPath[j->label] = next;
			next = j->label;
		}
	}

	for( const auto &i : modules ) {
		const string idString = i.first.toString();
		processFolder( sourceDir / idString, outputDir / idString, i.first, shimPath );
	}

}

static const fs::path &installedModulesPath() {
	static const fs::path s_installedModulesPath = AppData::dataDir() / "modules";
	return s_installedModulesPath;
}

static inline fs::path installedModulePath( const Uuid &moduleId ) {
	return installedModulesPath() / moduleId.toString();
}

fs::path AsmModuleStore::installedModulePath( const Uuid &moduleId, const Version &version ) {
	return installedModulePath( moduleId ) / version.toString();
}

static HashMap<Uuid,std::map<Version,AsmModuleDefinition>> s_store;

void AsmModuleStore::reload() {
	s_store.clear();
	fs::create_directories( installedModulesPath() );
	for( const auto &i : fs::directory_iterator( installedModulesPath() ) ) {
		if( i.status().type() != fs::file_type::directory ) continue;
		try {
			const Uuid moduleId = Uuid::parse( i.path().filename().u8string() );
			for( const auto &j : fs::directory_iterator( i.path() ) ) {
				if( j.status().type() != fs::file_type::directory || !fs::exists( j.path() / "module.json" ) ) continue;
				const std::optional<Version> version = Version::tryParse( j.path().filename().u8string() );
				if( !version.has_value() ) continue;
				try {
					AsmModuleDefinition module = AsmModuleDefinition::loadFromFile( (j.path() / "module.json").u8string() );
					if( module.version != version ) {
						std::cerr << "ASM Module version mismatch: " << (j.path() / "module.json").u8string() << std::endl;
						continue;
					}
					s_store[moduleId][version.value()] = std::move( module );
				} catch( ... ) {
					std::cerr << "Failed to parse installed ASM module: " << (j.path() / "module.json").u8string() << std::endl;
				}
			}
		} catch( ... ) {}
	}
}

const HashMap<Uuid,std::map<Version,AsmModuleDefinition>> &AsmModuleStore::allModules() {
	return s_store;
}

const std::map<Version,AsmModuleDefinition> &AsmModuleStore::getModuleVersions( const Uuid &moduleId ) {
	static const std::map<Version,AsmModuleDefinition> s_noResults;

	if( const auto versions = Util::tryGetValue( s_store, moduleId ) ) {
		return *versions;
	} else return s_noResults;
}

const AsmModuleDefinition &AsmModuleStore::installModule( const fs::path &modulePath ) {
	AsmModuleDefinition module = AsmModuleDefinition::loadFromFile( modulePath.u8string() );
	const fs::path installPath = installedModulePath( module.id, module.version );
	try {
		fs::forceDeleteRecursive( installPath );
		fs::create_directories( installPath.parent_path() );
		fs::copy( modulePath.parent_path(), installPath, fs::copy_options::recursive );
		clearReadOnlyFlagOnWindows( installPath );
	} catch( ... ) {
		if( auto versions = Util::tryGetValue( s_store, module.id ) ) {
			versions->erase( module.version );
		}
		throw;
	}
	AsmModuleDefinition &newModule = s_store[module.id][module.version];
	newModule = std::move( module );
	return newModule;
}

void AsmModuleStore::uninstallModule( const Uuid &moduleId ) {
	s_store.erase( moduleId );
	fs::forceDeleteRecursive( installedModulePath( moduleId ) );
}

void AsmModuleStore::uninstallModule( const Uuid &moduleId, const Version &version ) {
	s_store[moduleId].erase( version );
	if( s_store[moduleId].empty() ) {
		s_store.erase( moduleId );
		fs::forceDeleteRecursive( installedModulePath( moduleId ) );
	} else {
		fs::forceDeleteRecursive( installedModulePath( moduleId, version ) );
	}
}

struct BuiltinModule {
	Uuid id;
	Version version;
	const char *path;
};

void AsmModuleStore::installBuiltinModules() {
	const std::vector<BuiltinModule> builtinModules = {
		{ Uuid::parse( "8043cc1f-7789-44e2-99b4-ebb792fb2de4" ), Version{ 2, 0, 2 }, ":/modules/Pseudoinstructions_v2.0.2.bbpmod" },
		{ Uuid::parse( "32d81cb1-6499-42bd-bd75-be8f5360d857" ), Version{ 6, 0, 2 }, ":/modules/Helper_Functions_v6.0.2.bbpmod" },
		{ Uuid::parse( "49284c42-aeba-4e5d-a246-52af6debfaf3" ), Version{ 2, 1, 2 }, ":/modules/Wallrun_v2.1.2.bbpmod" },
		{ Uuid::parse( "9f5317da-4fb6-45b8-9267-ec1bc50d9aad" ), Version{ 1, 0, 4 }, ":/modules/Respawn_on_Void_Out_v1.0.4.bbpmod" },
		{ Uuid::parse( "acf89963-d184-42ba-a347-36c3d3f3dd2c" ), Version{ 1, 0, 2 }, ":/modules/2D_Area_v1.0.2.bbpmod" },
		{ Uuid::parse( "c3d9c5b3-c959-418b-bd1f-1d6d2859daf5" ), Version{ 1, 0, 2 }, ":/modules/Pause_Star_Hints_v1.0.2.bbpmod" },
		{ Uuid::parse( "8e976b20-7483-4d4d-9e4c-c902cef39f90" ), Version{ 1, 0, 2 }, ":/modules/Separate_Health_and_Air_v1.0.2.bbpmod" },
		{ Uuid::parse( "b6a6430c-47a9-4f61-b929-4a568d7be2ce" ), Version{ 1, 0, 2 }, ":/modules/Vertical_Intertia_v1.0.2.bbpmod" },
		{ Uuid::parse( "039954dd-47cd-4b9a-871c-c5726bfffd50" ), Version{ 1, 0, 2 }, ":/modules/True_Fixed_Camera_v1.0.2.bbpmod" },
		{ Uuid::parse( "bf8c9356-2140-424f-a861-fefdd9f0ef31" ), Version{ 1, 1, 0 }, ":/modules/Object_Camera_v1.1.0.bbpmod" },
		{ Uuid::parse( "ec56886a-edbc-4b76-81ba-9ff925bacee6" ), Version{ 1, 0, 1 }, ":/modules/Mosaic_Transitions_v1.0.1.bbpmod" }
	};

	reload();
	for( const BuiltinModule &module : builtinModules ) {
		if( !hasModuleVersion( module.id, module.version ) ) {
			//const fs::path tempDir =
			const fs::path tempDir = fs::to_path( createTemporaryDirectory() );
			const fs::path archivePath = tempDir / "module.bbpmod";
			const fs::path extractedPath = tempDir / "extracted";
			fs::create_directories( extractedPath );

			QFile::copy( module.path, archivePath.u8string().c_str() );
			clearReadOnlyFlagOnWindows( archivePath );
			Package::extract( archivePath, extractedPath );
			clearReadOnlyFlagOnWindows( extractedPath );

			installModule( extractedPath / "module.json" );
			fs::forceDeleteRecursive( tempDir );
		}
	}
}
