#include "src/core/data/behaviour.hpp"

#include "src/core/serialization-helpers.hpp"

const SimpleBehaviourMovement &SimpleBehaviourMovement::Default() {
	static const SimpleBehaviourMovement s_default = {
		/* type */ SimpleBehaviourInstruction::Move,
		/* frames */ 30,
		/* speed */ Vec3f{ 0, 0, 0 },
		/* angularVelocity */ Angle3s{ 0, 0, 0 }
	};
	return s_default;
};

const SimpleBehaviour &SimpleBehaviour::Default() {
	static const SimpleBehaviour s_default = {
		/* segmentedAddress */ 0,
		/* collisionDistanceMode */ CollisionDistanceMode::Suggested,
		/* renderDistance */ -1,
		/* collisionDistance */ -1,
		/* movesMario */ true,
		/* resetAfterLoop */ false,
		/* movementLoop */ std::vector<SimpleBehaviourMovement>()
	};
	return s_default;
};

static constexpr char P_TYPE[] = "type";

static constexpr char P_FRAMES[] = "frames";
static constexpr char P_SPEED_X[] = "speed_x";
static constexpr char P_SPEED_Y[] = "speed_y";
static constexpr char P_SPEED_Z[] = "speed_z";
static constexpr char P_SPEED_PITCH[] = "speed_pitch";
static constexpr char P_SPEED_YAW[] = "speed_yaw";
static constexpr char P_SPEED_ROLL[] = "speed_roll";

static constexpr char P_ROM_ADDR[] = "rom_address";
static constexpr char P_SEG_ADDR[] = "segmented_address";
static constexpr char P_COL_MODE[] = "collision_mode";
static constexpr char P_COL_DIST[] = "collision_distance";
static constexpr char P_DRAW_DIST[] = "render_distance";
static constexpr char P_INERTIA[] = "moves_mario";
static constexpr char P_RESETS[] = "resets";
static constexpr char P_LOOP[] = "loop";

template<> void JsonSerializer::serialize<SimpleBehaviourMovement>( JsonWriter &jw, const SimpleBehaviourMovement &obj ) {
	jw.writeObjectStart();
	jw.writeProperty( P_TYPE, obj.type );
	jw.writeProperty( P_FRAMES, obj.frames );
	if( (int)obj.type & (int)SimpleBehaviourInstruction::Move ) {
		jw.writeProperty( P_SPEED_X, obj.speed.x );
		jw.writeProperty( P_SPEED_Y, obj.speed.y );
		jw.writeProperty( P_SPEED_Z, obj.speed.z );
	}
	if( (int)obj.type & (int)SimpleBehaviourInstruction::Rotate ) {
		jw.writeProperty( P_SPEED_PITCH, obj.angularVelocity.pitch );
		jw.writeProperty( P_SPEED_YAW, obj.angularVelocity.yaw );
		jw.writeProperty( P_SPEED_ROLL, obj.angularVelocity.roll );
	}
	jw.writeObjectEnd();
}

template<> SimpleBehaviourMovement JsonSerializer::parse<SimpleBehaviourMovement>( const Json &json ) {
	return SimpleBehaviourMovement {
		json[P_TYPE].get<SimpleBehaviourInstruction>(),
		json[P_FRAMES].get<short>(),
		Vec3f{
			json[P_SPEED_X].getOrDefault<float>( 0.f ),
			json[P_SPEED_Y].getOrDefault<float>( 0.f ),
			json[P_SPEED_Z].getOrDefault<float>( 0.f )
		},
		Angle3s{
			json[P_SPEED_PITCH].getOrDefault<short>( 0 ),
			json[P_SPEED_YAW].getOrDefault<short>( 0 ),
			json[P_SPEED_ROLL].getOrDefault<short>( 0 )
		}
	};
}

template<> void JsonSerializer::serialize<Behaviour>( JsonWriter &jw, const Behaviour &obj ) {
	jw.writeObjectStart();
	jw.writeProperty( P_TYPE, obj.type );
	switch( obj.type ) {
		case BehaviourType::Vanilla:
			jw.writeProperty( P_ROM_ADDR, obj.romAddress );
			break;
		case BehaviourType::Autogenerated:
			jw.writeProperty( P_SEG_ADDR, obj.autoGen.segmentedAddress );
			jw.writeProperty( P_COL_MODE, obj.autoGen.collisionDistanceMode );
			jw.writeProperty( P_COL_DIST, obj.autoGen.collisionDistance );
			jw.writeProperty( P_DRAW_DIST, obj.autoGen.renderDistance );
			jw.writeProperty( P_INERTIA, obj.autoGen.movesMario );
			jw.writeProperty( P_RESETS, obj.autoGen.resetAfterLoop );
			serializeIterator( jw, P_LOOP, obj.autoGen.movementLoop.begin(), obj.autoGen.movementLoop.end() );
			break;
		default: break;
	}
	jw.writeObjectEnd();
}

template<> Behaviour JsonSerializer::parse<Behaviour>( const Json &json ) {
	if( json["collision_asm_ref"].exists() ) {
		// Migrate from old schema
		if( json["collision_asm_ref"].get<bool>() ) {
			return { BehaviourType::AsmReference, 0, SimpleBehaviour::Default() };
		} else {
			return { BehaviourType::Vanilla, json["collision_ptr_loc"].get<uint>(), SimpleBehaviour::Default() };
		}
	}

	Behaviour beh = {
		json[P_TYPE].get<BehaviourType>(),
		0,
		SimpleBehaviour::Default()
	};

	switch( beh.type ) {
		case BehaviourType::Vanilla:
			beh.romAddress = json[P_ROM_ADDR].get<uint>();
			break;
		case BehaviourType::Autogenerated: {
			beh.autoGen = {
				json[P_SEG_ADDR].get<uint>(),
				json[P_COL_MODE].get<CollisionDistanceMode>(),
				json[P_DRAW_DIST].get<short>(),
				json[P_COL_DIST].get<short>(),
				json[P_INERTIA].get<bool>(),
				json[P_RESETS].get<bool>(),
				parseVector<SimpleBehaviourMovement>( json[P_LOOP] )
			};
			break;
		}
		default: break;
	}

	return beh;
}



