#include "src/core/sound.hpp"

#include <cstdio>

enum PlaybackFlags : ubyte {
	LOOP = 0x0,
	FADE = 0x1,
	NO_ECHO = 0x2,
	NO_LOOP = 0x8
};

enum SoundFlags : ubyte {
	NONE = 0x0,
	CONSTANT_VOLUME = 0x1, // Volume not affected by distance
	VIBRATO = 0x2, // Audio frequency scale is quasi-randomly set each audio frame
	CONSTANT_PRIORITY = 0x4, // Priority is not affected by distance
	NO_ATTENUATION = 0x8 // Frequency is not affected by distance. If in bank 1, it's volume and left/right balance is also fixed
};

static inline PlaybackFlags operator|( PlaybackFlags f1, PlaybackFlags f2 ) {
	return (PlaybackFlags)((ubyte)f1 | (ubyte)f2 );
}

static inline SoundFlags operator|( SoundFlags f1, SoundFlags f2 ) {
	return (SoundFlags)((ubyte)f1 | (ubyte)f2 );
}

struct SoundInfo {
	const char *name;
	SoundFlags upperFlags;
	ubyte priority;
	PlaybackFlags lowerFlags;
};

static const SoundInfo UNDEFINED{ nullptr, NONE, 0, LOOP };

static const SoundInfo bank0[] = {
	/* 0x00 */	{ "Mario Jump",							CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x01 */	{ "Mario Jump (Grass)",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x02 */	{ "Mario Jump (Water)",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x03 */	{ "Mario Jump (Stone)",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x04 */	{ "Mario Jump (Spooky)",                CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x05 */	{ "Mario Jump (Snow)",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x06 */	{ "Mario Jump (Ice)",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x07 */	{ "Mario Jump (Sand)",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x08 */	{ "Mario Landed",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x09 */	{ "Mario Landed (Grass)",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x0A */	{ "Mario Landed (Water)",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x0B */	{ "Mario Landed (Stone)",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x0C */	{ "Mario Landed (Spooky)",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x0D */	{ "Mario Landed (Snow)",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x0E */	{ "Mario Landed (Ice)",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x0F */	{ "Mario Landed (Sand)",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x10 */	{ "Mario Step",							CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x11 */	{ "Mario Step (Grass)",					CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x12 */	{ "Mario Step (Water)",					CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x13 */	{ "Mario Step (Stone)",					CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x14 */	{ "Mario Step (Spooky)",				CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x15 */	{ "Mario Step (Snow)",					CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x16 */	{ "Mario Step (Ice)",					CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x17 */	{ "Mario Step (Sand)",					CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x18 */	{ "Mario Fell Down",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x19 */	{ "Mario Fell Down (Grass)",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x1A */	{ "Mario Fell Down (Water)",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x1B */	{ "Mario Fell Down (Stone)",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x1C */	{ "Mario Fell Down (Spooky)",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x1D */	{ "Mario Fell Down (Snow)",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x1E */	{ "Mario Fell Down (Ice)",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x1F */	{ "Mario Fell Down (Sand)",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x20 */	{ "Mario Tiptoe",						CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x21 */	{ "Mario Tiptoe (Grass)",				CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x22 */	{ "Mario Tiptoe (Water)",				CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x23 */	{ "Mario Tiptoe (Stone)",				CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x24 */	{ "Mario Tiptoe (Spooky)",				CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x25 */	{ "Mario Tiptoe (Snow)",				CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x26 */	{ "Mario Tiptoe (Ice)",					CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x27 */	{ "Mario Tiptoe (Sand)",				CONSTANT_PRIORITY | VIBRATO, 0x80, NO_LOOP },
	/* 0x28 */	{ "Metal Mario Jump",					CONSTANT_PRIORITY, 0x90, NO_LOOP },
	/* 0x29 */	{ "Metal Mario Landed",					CONSTANT_PRIORITY, 0x90, NO_LOOP },
	/* 0x2A */	{ "Metal Mario Step",					CONSTANT_PRIORITY, 0x90, NO_LOOP },
	/* 0x2B */	{ "Metal Mario Hard Landing",			CONSTANT_PRIORITY, 0x90, NO_LOOP },
	/* 0x2C */	{ "Mario Clap",							CONSTANT_PRIORITY | VIBRATO, 0x00, NO_LOOP },
	/* 0x2D */	{ "Mario Hanging",						CONSTANT_PRIORITY, 0xA0, NO_LOOP },
	/* 0x2E */	{ "Mario Step (Quicksand)",				CONSTANT_PRIORITY, 0x00, NO_LOOP },
	/* 0x2F */	{ "Metal Mario Tiptoe",					CONSTANT_PRIORITY, 0x90, NO_LOOP },
	/* 0x30 */	{ "Water Splash",						CONSTANT_PRIORITY, 0xC0, NO_LOOP },
	/* 0x31 */	{ "Tiny Water Splash",					CONSTANT_PRIORITY, 0x60, NO_LOOP },
	/* 0x32 */	{ "Quiet Water Splash",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x33 */	{ "Mario Swimming",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x34 */	{ "Water Bubbles",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x35 */	{ "Mario Throwing Item",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x36 */	{ "Key Swish",							CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x37 */	{ "Mario Spin",							CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x38 */	{ "Mario Twirl",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x39 */	{ "Mario Twirl",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x3A */	{ "Mario Climb Up Tree",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x3B */	{ "Mario Climb Down Tree",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x3C */	{ "Mario Step: Unknown 1",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x3D */	{ "Mario Step: Unknown 2",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x3E */	{ "Mario Step: Unknown 3",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x3F */	{ "Mario Pats Back",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x40 */	{ "Mario Brush Hair",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x41 */	{ "Mario Climb Up Pole",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x42 */	{ "Metal Mario Bonk",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x43 */	{ "Mario Pops Out of Ground",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x44 */	{ "Mario Hit 1",						CONSTANT_PRIORITY, 0xA0, NO_LOOP },
	/* 0x45 */	{ "Mario Bonk",							CONSTANT_PRIORITY, 0xA0, NO_LOOP },
	/* 0x46 */	{ "Mario Shrink Into BBH",				CONSTANT_PRIORITY, 0xA0, NO_LOOP },
	/* 0x47 */	{ "Mario Swimming Quickly",				CONSTANT_PRIORITY, 0xA0, NO_LOOP },
	/* 0x48 */	{ "Mario Stuck In Ground",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x49 */	{ "Mario Stuck In Ground (Grass)",		CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x4A */	{ "Mario Stuck In Ground (Water)",		CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x4B */	{ "Mario Stuck In Ground (Stone)",		CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x4C */	{ "Mario Stuck In Ground (Spooky)",		CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x4D */	{ "Mario Stuck In Ground (Snow)",		CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x4E */	{ "Mario Stuck In Ground (Ice)",		CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x4F */	{ "Mario Stuck In Ground (Sand)",		CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x50 */	{ "Metal Mario Jump in Water",			CONSTANT_PRIORITY, 0x90, NO_LOOP },
	/* 0x51 */	{ "Metal Mario Land in Water",			CONSTANT_PRIORITY, 0x90, NO_LOOP },
	/* 0x52 */	{ "Metal Mario Step in Water",			CONSTANT_PRIORITY, 0x90, NO_LOOP },
	/* 0x53 */	{ "Quiet Metal Pound",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x54 */	{ "Unused Quiet Clunk",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x55 */	{ "Unused Sound Effect: Bweeew",		CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x56 */	{ "Flying Fast",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x57 */	{ "Teleport",							CONSTANT_PRIORITY, 0xC0, NO_LOOP },
	/* 0x58 */	{ "Chink! (Bullet Hits Metal Mario)",	CONSTANT_PRIORITY, 0xA0, NO_LOOP },
	/* 0x59 */	{ "Mario Bonk Off Object",				CONSTANT_PRIORITY, 0xB0, NO_LOOP },
	/* 0x5A */	{ "Mario Side Flip",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x5B */	{ "Mario Read Sign",					CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x5C */	{ "Unused Double Step",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x5D */	{ "Unused Bonk Sound",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x5E */	{ "Lakitu Zoom",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x5F */	{ "Lakitu Zoom (2)",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x60 */	{ "Mario Hard Landing",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x61 */	{ "Mario Hard Landing (Grass)",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x62 */	{ "Mario Hard Landing (Water)",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x63 */	{ "Mario Hard Landing (Stone)",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x64 */	{ "Mario Hard Landing (Spooky)",		CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x65 */	{ "Mario Hard Landing (Snow)",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x66 */	{ "Mario Hard Landing (Ice)",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x67 */	{ "Mario Hard Landing (Sand)",			CONSTANT_PRIORITY, 0x80, NO_LOOP }
};

static const SoundInfo bank1[] = {
	/* 0x00 */	{ "Mario Sliding",						CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x01 */	{ "Mario Sliding (Grass)",				CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x02 */	{ "Mario Sliding (Water)",				CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x03 */	{ "Mario Sliding (Stone)",				CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x04 */	{ "Mario Sliding (Spooky)",				CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x05 */	{ "Mario Sliding (Snow)",				CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x06 */	{ "Mario Sliding (Ice)",				CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x07 */	{ "Mario Sliding (Sand)",				CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x08 */	UNDEFINED,
	/* 0x09 */	UNDEFINED,
	/* 0x0A */	UNDEFINED,
	/* 0x0B */	UNDEFINED,
	/* 0x0C */	UNDEFINED,
	/* 0x0D */	UNDEFINED,
	/* 0x0E */	UNDEFINED,
	/* 0x0F */	UNDEFINED,
	/* 0x10 */	{ "Burning",							CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x11 */	{ "Slide Down Pole",					CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x12 */	{ "Slide Down Tree",					CONSTANT_PRIORITY, 0x80, LOOP },
	/* 0x13 */	{ "Bling",								CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x14 */	{ "Quicksand Death",					CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x15 */	{ "Bling (2)",							CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x16 */	{ "Shocked",							CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x17 */	{ "Mario Flying",						CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x18 */	{ "Mario Drowning",						NO_ATTENUATION | CONSTANT_PRIORITY, 0x00, LOOP },
	/* 0x19 */	{ "Aiming Cannon",						NO_ATTENUATION | CONSTANT_PRIORITY | CONSTANT_VOLUME, 0x20, LOOP },
	/* 0x1A */	{ "Unused Annoying Alarm",				CONSTANT_PRIORITY, 0x80, LOOP },
	/* 0x1B */	UNDEFINED,
	/* 0x1C */	UNDEFINED,
	/* 0x1D */	UNDEFINED,
	/* 0x1E */	UNDEFINED,
	/* 0x1F */	UNDEFINED,
	/* 0x20 */	{ "Riding Shell",						CONSTANT_PRIORITY, 0x20, LOOP },
	/* 0x21 */	{ "Riding Shell (Grass)",				CONSTANT_PRIORITY, 0x20, LOOP },
	/* 0x22 */	{ "Riding Shell (Water)",				CONSTANT_PRIORITY, 0x20, LOOP },
	/* 0x23 */	{ "Riding Shell (Stone)",				CONSTANT_PRIORITY, 0x20, LOOP },
	/* 0x24 */	{ "Riding Shell (Spooky)",				CONSTANT_PRIORITY, 0x20, LOOP },
	/* 0x25 */	{ "Riding Shell (Snow)",				CONSTANT_PRIORITY, 0x20, LOOP },
	/* 0x26 */	{ "Riding Shell (Ice)",					CONSTANT_PRIORITY, 0x20, LOOP },
	/* 0x27 */	{ "Riding Shell (Sand)",				CONSTANT_PRIORITY, 0x20, LOOP },
	/* 0x28 */	{ "Riding Shell (Lava)",				CONSTANT_PRIORITY, 0x28, LOOP }
};

static const SoundInfo bank2[] = {
	/* 0x00 */	{ "Mario Sound: Yah!",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x01 */	{ "Mario Sound: Wah!",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x02 */	{ "Mario Sound: Hoo!",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x03 */	{ "Mario Sound: Hoohoo!",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x04 */	{ "Mario Sound: Yahoo! (2)",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x05 */	{ "Mario Sound: Uh",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x06 */	{ "Mario Sound: Hmm",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x07 */	{ "Mario Sound: Wah! (2)",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x08 */	{ "Mario Sound: Woah!",					CONSTANT_PRIORITY, 0xC0, NO_LOOP },
	/* 0x09 */	{ "Mario Sound: Eeungh",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x0A */	{ "Mario Sound: Oooough!",				CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x0B */	{ "Mario Sound: Oof!",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x0C */	{ "Mario Sound: Oof! (2)",				CONSTANT_PRIORITY, 0xD0, NO_LOOP },
	/* 0x0D */	{ "Mario Sound: Here We Go!",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x0E */	{ "Mario Yawning",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x0F */	{ "Mario Snoring",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x10 */	{ "Mario Snoring (2)",					CONSTANT_PRIORITY, 0xC0, NO_LOOP },
	/* 0x11 */	{ "Mario Sound: Waaaaoooooooowwwww",	CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x12 */	{ "Mario Sound: Haha!",					CONSTANT_PRIORITY, 0xF0, NO_LOOP },
	/* 0x13 */	{ "Mario Sound: Whoo (Low Health)",		CONSTANT_PRIORITY, 0xD0, NO_LOOP },
	/* 0x14 */	{ "Mario Sound: Uh (2)",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x15 */	{ "Mario Sound: Hahahahahahahahahaaaa",	CONSTANT_PRIORITY, 0xA0, NO_LOOP },
	/* 0x16 */	{ "Mario Sound: Eeeeuuuuggghhhh",		CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x17 */	{ "Mario Sound: Whoo (Cold)",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x18 */	{ "Mario Panting",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x19 */	{ "Mario Panting (2)",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x1A */	{ "Mario Panting (3)",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x1B */	{ "Mario Coughing",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x1C */	{ "Mario Coughing (2)",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x1D */	{ "Mario Coughing (3)",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x1E */	{ "Mario Punch 1",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x1F */	{ "Mario Punch 3",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x20 */	{ "Mario Sound: Mama Mia",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x21 */	{ "Mario Sound: Okie Dokie",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x22 */	{ "Mario Sound: Wah (Ground Pound)",	CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x23 */	{ "Mario Drowned",						CONSTANT_PRIORITY, 0xF0, NO_LOOP },
	/* 0x24 */	{ "Mario Punch 2",						CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x25 */	{ "Mario Sound: Unf",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x26 */	{ "Mario Sound: Hnnnnggg",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x27 */	{ "Mario Sound: Yah! (Distorted)",		CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x28 */	{ "Peach: Dear Mario",					CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x29 */	UNDEFINED,
	/* 0x2A */	UNDEFINED,
	/* 0x2B */	{ "Mario Sound: Yahoo!",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x2C */	UNDEFINED,
	/* 0x2D */	UNDEFINED,
	/* 0x2E */	{ "Mario Sound: Waha!",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x2F */	{ "Mario Sound: Yippee!",				CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x30 */	{ "Mario Sound: Doh!",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x31 */	{ "Mario Sound: Game Over",				CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x32 */	{ "Mario Sound: Hello",					CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x33 */	{ "Mario Sound: Press Start to Play",	CONSTANT_PRIORITY, 0xFF, NO_LOOP | NO_ECHO },
	/* 0x34 */	{ "Mario Twirl Bounce",					CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x35 */	{ "Mario Snoring (3)",					CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x36 */	{ "Mario Sound: So Long Eh Bowser",		CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x37 */	{ "Mario Sound: I'm'a Tired",			CONSTANT_PRIORITY, 0x80, NO_LOOP },
	/* 0x38 */	{ "Peach: Mario",						CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x39 */	{ "Peach: The Power of the Stars...",	CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x3A */	{ "Peach: It's All Thanks to You",		CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x3B */	{ "Peach: Thank You, Mario",			CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x3C */	{ "Peach: We Need to Do Something...",	CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x3D */	{ "Peach: Listen, Everybody...",		CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x3E */	{ "Peach: For Mario",					CONSTANT_PRIORITY, 0xFF, NO_LOOP },
	/* 0x3F */	{ "Peach: Mariooo",						CONSTANT_PRIORITY, 0xFF, NO_LOOP }
};

static const SoundInfo bank3[] = {
	/* 0x00 */	{ "Cap Switch Activated",				NONE, 0x80, NO_LOOP },
	/* 0x01 */	UNDEFINED,
	/* 0x02 */	UNDEFINED,
	/* 0x03 */	{ "Flame Fizzles Out",					NONE, 0x80, NO_LOOP },
	/* 0x04 */	{ "Wooden Door Opening",				NONE, 0xC0, NO_LOOP },
	/* 0x05 */	{ "Wooden Door Closing",				NONE, 0xC0, NO_LOOP },
	/* 0x06 */	{ "Iron Door Opening",					NONE, 0xC0, NO_LOOP },
	/* 0x07 */	{ "Iron Door Closing",					NONE, 0xC0, NO_LOOP },
	/* 0x08 */	{ "Bubbles",							NONE, 0x00, NO_LOOP },
	/* 0x09 */	{ "Flowing Water",						NONE, 0x00, NO_LOOP },
	/* 0x0A */	{ "Water Swish",						NONE, 0x00, NO_LOOP },
	/* 0x0B */	{ "Quiet Bubble",						NONE, 0x00, NO_LOOP },
	/* 0x0C */	{ "Volcano Explosion",					NONE, 0x80, NO_LOOP },
	/* 0x0D */	{ "Quiet Bubble (2)",					NONE, 0x00, NO_LOOP },
	/* 0x0E */	{ "Trapdoor Opening",					NONE, 0x80, NO_LOOP },
	/* 0x0F */	{ "WF Wall Broken",						NONE, 0x00, NO_LOOP },
	/* 0x10 */	UNDEFINED,
	/* 0x11 */	{ "Coin Collected",						NO_ATTENUATION, 0x80, NO_LOOP },
	/* 0x12 */	{ "Coin Collected Underwater",			NO_ATTENUATION, 0x80, NO_LOOP },
	/* 0x13 */	UNDEFINED,
	/* 0x14 */	UNDEFINED,
	/* 0x15 */	UNDEFINED,
	/* 0x16 */	{ "Present Star (Short)",				NONE, 0x00, NO_LOOP | FADE },
	/* 0x17 */	{ "TTC Clock Hand Tick",				NONE, 0x00, NO_LOOP },
	/* 0x18 */	{ "Loud Pound",							NONE, 0x80, NO_LOOP },
	/* 0x19 */	{ "Loud Pound (2)",						NONE, 0x80, NO_LOOP },
	/* 0x1A */	{ "Short Pound",						NONE, 0x80, NO_LOOP },
	/* 0x1B */	{ "Short Pound (2)",					NONE, 0x80, NO_LOOP },
	/* 0x1C */	{ "Short Pound (3)",					NONE, 0x80, NO_LOOP },
	/* 0x1D */	{ "Short Pound (4)",					NONE, 0x80, NO_LOOP },
	/* 0x1E */	{ "Short Pound (5)",					NONE, 0x80, NO_LOOP },
	/* 0x1F */	{ "Short Pound (6)",					NONE, 0x80, NO_LOOP },
	/* 0x20 */	{ "Chest Opened",						CONSTANT_VOLUME, 0x80, NO_LOOP },
	/* 0x21 */	UNDEFINED,
	/* 0x22 */	{ "Clam Shell",							CONSTANT_VOLUME, 0x80, NO_LOOP },
	/* 0x23 */	UNDEFINED,
	/* 0x24 */	{ "Throwable Box Landing",				NONE, 0x00, NO_LOOP },
	/* 0x25 */	{ "Rolling Log Turning",				NONE, 0x00, NO_LOOP },
	/* 0x26 */	{ "Clam Shell (2)",						NONE, 0x40, NO_LOOP },
	/* 0x27 */	{ "Clam Shell (3)",						NONE, 0x40, NO_LOOP },
	/* 0x28 */	{ "Exit Painting",						NO_ATTENUATION | CONSTANT_VOLUME, 0x00, NO_LOOP },
	/* 0x29 */	UNDEFINED,
	/* 0x2A */	UNDEFINED,
	/* 0x2B */	{ "Level Select Change Stage",			NONE, 0x00, NO_LOOP },
	/* 0x2C */	UNDEFINED,
	/* 0x2D */	{ "Bridge Piece Falling",				NONE, 0x80, NO_LOOP },
	/* 0x2E */	{ "Bob-Omb Explosion",					NONE, 0x20, NO_LOOP },
	/* 0x2F */	{ "Bowser Bomb Explosion",				CONSTANT_VOLUME, 0x00, NO_LOOP },
	/* 0x30 */	{ "Coin Spurt",							NONE, 0x00, NO_LOOP },
	/* 0x31 */	{ "Explosion",						NONE, 0x80, NO_LOOP },
	/* 0x32 */	UNDEFINED,
	/* 0x33 */	UNDEFINED,
	/* 0x34 */	{ "Boat Tilt Up",						NONE, 0x40, NO_LOOP },
	/* 0x35 */	{ "Boat Tilt Down",						NONE, 0x40, NO_LOOP },
	/* 0x36 */	{ "Coin Spawn",							NONE, 0x40, NO_LOOP },
	/* 0x37 */	{ "Key Bounce",							NONE, 0x00, NO_LOOP },
	/* 0x38 */	{ "TTC Pendulum",						NONE, 0x80, NO_LOOP },
	/* 0x39 */	{ "Chain Chomp Hop",					NONE, 0x00, NO_LOOP },
	/* 0x3A */	{ "Chain Chomp Lunge",					NONE, 0x00, NO_LOOP },
	/* 0x3B */	{ "Key Door - Key Turned",				NONE, 0x00, NO_LOOP },
	/* 0x3C */	{ "Moving in Sand",						NONE, 0x00, NO_LOOP },
	/* 0x3D */	{ "Generic Clunk",						NONE, 0x00, NO_LOOP },
	/* 0x3E */	{ "Moving Platform Switch",				NONE, 0x00, NO_LOOP },
	/* 0x3F */	{ "Cage Opened",						NONE, 0xA0, NO_LOOP },
	/* 0x40 */	{ "Quiet Pound",						NONE, 0x40, NO_LOOP },
	/* 0x41 */	{ "Box Break",							NONE, 0xC0, NO_LOOP },
	/* 0x42 */	{ "Key Door - Key Inserted",			NONE, 0x00, NO_LOOP },
	/* 0x43 */	{ "Quiet Pound (2)",					NONE, 0x00, NO_LOOP },
	/* 0x44 */	{ "Pillar Smashed",						NONE, 0x00, NO_LOOP },
	/* 0x45 */	{ "Super Quiet Rolling Log Sound",		NONE, 0x00, NO_LOOP },
	/* 0x46 */	{ "Tox Box Roll",						NONE, 0x00, NO_LOOP },
	/* 0x47 */	{ "Cannon Rising",						NONE, 0x80, NO_LOOP },
	/* 0x48 */	{ "Grindel Roll",						NONE, 0x80, NO_LOOP },
	/* 0x49 */	{ "Explosion (2)",						NONE, 0x80, NO_LOOP },
	/* 0x4A */	{ "Shaking Coffin",						NONE, 0x80, NO_LOOP },
	/* 0x4B */	{ "Pyramid Top Spinning",				CONSTANT_VOLUME, 0xE0, NO_LOOP },
	/* 0x4C */	{ "Pyramid Top Explosion",				CONSTANT_VOLUME, 0xF0, NO_LOOP },
	/* 0x4D */	{ "Starter Pistol",						CONSTANT_VOLUME, 0x40, NO_LOOP },
	/* 0x4E */	{ "Star Door Opening",					NONE, 0xC0, NO_LOOP },
	/* 0x4F */	{ "Star Door Closing",					NONE, 0xC0, NO_LOOP },
	/* 0x50 */	{ "Bird Chirp (1)",						NONE, 0x40, LOOP },
	/* 0x51 */	UNDEFINED,
	/* 0x52 */	UNDEFINED,
	/* 0x53 */	UNDEFINED,
	/* 0x54 */	{ "Switch Timer Ticking Fast",			NONE, 0xF0, FADE },
	/* 0x55 */	{ "Switch Timer Ticking Slow",			NONE, 0xF0, FADE },
	/* 0x56 */	{ "Falling Pillar Breaks",				NONE, 0x00, NO_LOOP },
	/* 0x57 */	{ "Star Spawn (2)",						NONE, 0xFF, NO_LOOP | FADE },
	/* 0x58 */	{ "1UP",								NONE, 0xFF, NO_LOOP },
	/* 0x59 */	{ "TTC Rotating Block Tick",			NONE, 0x00, NO_LOOP },
	/* 0x5A */	{ "Kickable Board Wobble",				NONE, 0x00, NO_LOOP },
	/* 0x5B */	{ "Elevator Moving",					NONE, 0x00, NO_LOOP },
	/* 0x5C */	{ "Haunted Chair Spinning",				NONE, 0x00, NO_LOOP },
	/* 0x5D */	{ "Haunted Chair Launch",				NONE, 0x00, NO_LOOP },
	/* 0x5E */	{ "Boo Cage Lands on Ground",			NONE, 0x00, NO_LOOP },
	/* 0x5F */	{ "Haunted Chair Rises",				NONE, 0x00, NO_LOOP },
	/* 0x60 */	{ "Bowser Explode",						NONE, 0x00, NO_LOOP },
	/* 0x61 */	{ "Key Spawn",							NONE, 0x00, NO_LOOP },
	/* 0x62 */	{ "Bowser Platform Falling",			NONE, 0x80, NO_LOOP },
	/* 0x63 */	{ "1UP Spawn",							NONE, 0xD0, NO_LOOP },
	/* 0x64 */	{ "Spinning Heart",						NONE, 0xC0, NO_LOOP },
	/* 0x65 */	{ "Wooden Post Pounded",				NONE, 0xC0, NO_LOOP },
	/* 0x66 */	{ "Water Level Changed",				NONE, 0x80, NO_LOOP },
	/* 0x67 */	{ "Blue Coin Switch Activated",			NONE, 0xA0, NO_LOOP },
	/* 0x68 */	{ "Red Coin (JP Version)",				NONE, 0x90, NO_LOOP },
	/* 0x69 */	{ "Birds Flying Away",					NONE, 0x00, NO_LOOP },
	/* 0x6A */	{ "Correct Chest Chosen",				NONE, 0xA0, NO_LOOP },
	/* 0x6B */	{ "Elevator Reached Bottom",			NONE, 0x80, NO_LOOP },
	/* 0x6C */	{ "Bounce Low",							NONE, 0x40, NO_LOOP },
	/* 0x6D */	{ "Bounce High",						NONE, 0x40, NO_LOOP },
	/* 0x6E */	{ "Yoshi Step",							NONE, 0x20, NO_LOOP },
	/* 0x6F */	{ "Yoshi Jump",							NONE, 0x30, NO_LOOP },
	/* 0x70 */	{ "Yoshi Talk",							NONE, 0x30, NO_LOOP },
	/* 0x71 */	{ "Splatter",							NONE, 0x30, NO_LOOP },
	/* 0x72 */	{ "Beta Trampoline?",					NONE, 0x80, NO_LOOP },
	/* 0x73 */	{ "Grand Star",							NONE, 0x00, NO_LOOP },
	/* 0x74 */	{ "Grand Star Bounce",					NONE, 0x00, NO_LOOP },
	/* 0x75 */	{ "RR Boat Rowing (JP Version)",		NONE, 0x00, NO_LOOP },
	/* 0x76 */	{ "Moneybag Reveal",					NONE, 0x20, NO_LOOP }
};

static const SoundInfo bank4[] = {
	/* 0x00 */	{ "Waterfall",							NONE, 0x00, LOOP },
	/* 0x01 */	{ "Waterfall (2)",						NONE, 0x00, LOOP },
	/* 0x02 */	{ "Elevator",							NONE, 0x00, LOOP },
	/* 0x03 */	{ "Droning",							CONSTANT_VOLUME, 0x00, LOOP },
	/* 0x04 */	{ "Droning (2)",						NONE, 0x00, LOOP },
	/* 0x05 */	{ "Wind",								NONE, 0x00, LOOP },
	/* 0x06 */	{ "Moving Through Sand/Snow",			NONE, 0x00, LOOP },
	/* 0x07 */	{ "Unused Stone Dragging Sound",		NONE, 0x00, LOOP },
	/* 0x08 */	{ "Quiet Hammering",					NONE, 0x00, LOOP },
	/* 0x09 */	{ "Water Sloshing",						NONE, 0x00, LOOP },
	/* 0x0A */	{ "Rolling Ball",						NONE, 0x00, LOOP },
	/* 0x0B */	{ "Boat Rocking",						NONE, 0x00, LOOP },
	/* 0x0C */	{ "Moving Platform on Track",			NONE, 0x00, LOOP },
	/* 0x0D */	{ "Express Elevator",					CONSTANT_VOLUME, 0x00, LOOP },
	/* 0x0E */	{ "Moving Sand",						NONE, 0x00, LOOP },
	/* 0x0F */	{ "Merry-Go-Round Creaking",			NONE, 0x40, LOOP },
	/* 0x10 */	{ "Wind (2)",							NONE, 0x80, LOOP },
	/* 0x11 */	UNDEFINED,
	/* 0x12 */	{ "Unused Buzzing Sound",				NONE, 0x00, LOOP },
	/* 0x13 */	{ "Sliding",							NONE, 0x00, LOOP },
	/* 0x14 */	{ "Star Glitter",						NONE, 0x00, FADE },
	/* 0x15 */	{ "Bowser 2 Platform Tilted",			CONSTANT_VOLUME, 0x00, LOOP },
	/* 0x16 */	{ "Water Draining",						CONSTANT_VOLUME, 0x00, LOOP },
	/* 0x17 */	{ "Metal Box Pushed",					NONE, 0x80, LOOP },
	/* 0x18 */	{ "Sinking Quicksand",					NONE, 0x80, LOOP }
};

static const SoundInfo bank5[] = {
	/* 0x00 */	{ "Sushi Swimming",						NONE, 0x80, NO_LOOP },
	/* 0x01 */	{ "Mr. I Shoot",						NONE, 0x00, NO_LOOP },
	/* 0x02 */	{ "Baby Penguin Walk",					NONE, 0x00, NO_LOOP },
	/* 0x03 */	{ "Bowser Walk",						NONE, 0x00, NO_LOOP },
	/* 0x04 */	{ "Bowser Roar",						NONE, 0x00, NO_LOOP },
	/* 0x05 */	{ "Bowser's Tail Grabbed",				NONE, 0x00, NO_LOOP },
	/* 0x06 */	{ "Bowser Defeated",					NONE, 0x00, NO_LOOP },
	/* 0x07 */	{ "Bowser Spinning",					NONE, 0x00, NO_LOOP },
	/* 0x08 */	{ "Bowser Breath In",					NONE, 0x00, NO_LOOP },
	/* 0x09 */	{ "Mother Penguin Walk",				NONE, 0x80, NO_LOOP },
	/* 0x0A */	{ "Boo Bounce",							NONE, 0x00, NO_LOOP },
	/* 0x0B */	{ "Boo Laugh (Short)",					NONE, 0x00, NO_LOOP },
	/* 0x0C */	{ "Thwomp",								NONE, 0xA0, NO_LOOP },
	/* 0x0D */	{ "Enter Cannon 1",						NONE, 0xF0, NO_LOOP },
	/* 0x0E */	{ "Enter Cannon 2",						NONE, 0xF0, NO_LOOP },
	/* 0x0F */	{ "Enter Cannon 3",						NONE, 0xF0, NO_LOOP },
	/* 0x10 */	{ "Pirhana Plant Bite",					NONE, 0x50, NO_LOOP },
	/* 0x11 */	{ "Prihana Plant Defeated",				NONE, 0x60, NO_LOOP },
	/* 0x12 */	{ "Unused Water Slosh",					NONE, 0x80, NO_LOOP },
	/* 0x13 */	{ "Bomp Extend",						NONE, 0x00, NO_LOOP },
	/* 0x14 */	{ "Mr. I Defeated",						NONE, 0x00, NO_LOOP },
	/* 0x15 */	{ "Pounding",							NONE, 0x50, NO_LOOP },
	/* 0x16 */	{ "Boss Hurt",							NONE, 0x60, NO_LOOP },
	/* 0x17 */	{ "Bully Clank",						NONE, 0x80, NO_LOOP },
	/* 0x18 */	{ "Bully Defeated",						NONE, 0xA0, NO_LOOP },
	/* 0x19 */	{ "Bowser Puzzle Piece Moved",			NONE, 0x20, NO_LOOP },
	/* 0x1A */	{ "Bullet Bill Launched",				NONE, 0x50, NO_LOOP },
	/* 0x1B */	{ "Bully Walk",							NONE, 0x30, NO_LOOP },
	/* 0x1C */	{ "Bully Pushed Back",					NONE, 0x00, NO_LOOP },
	/* 0x1D */	{ "Mario Grabbed",						NONE, 0x80, NO_LOOP },
	/* 0x1E */	{ "Mario Thrown",						NONE, 0xA0, NO_LOOP },
	/* 0x1F */	{ "Baby Penguin Dive",					NONE, 0x40, NO_LOOP },
	/* 0x20 */	{ "Goomba Walk",						NONE, 0x00, NO_LOOP },
	/* 0x21 */	{ "Monkey Noises 3",					NONE, 0x00, NO_LOOP },
	/* 0x22 */	{ "Enemy Throw Projectile",				NONE, 0x00, NO_LOOP },
	/* 0x23 */	{ "Unused Pew Sound",					NONE, 0x80, NO_LOOP },
	/* 0x24 */	{ "Generic Enemy Death",				NONE, 0x40, NO_LOOP },
	/* 0x25 */	{ "Water Bomb Cannon Fired",			NONE, 0x40, NO_LOOP },
	/* 0x26 */	{ "Unused Wee-Wah Sound",				NONE, 0x80, NO_LOOP },
	/* 0x27 */	{ "Bob-Omb Walk",						NONE, 0x00, NO_LOOP },
	/* 0x28 */	{ "Water Bomb Falling",					NONE, 0x80, NO_LOOP },
	/* 0x29 */	{ "Water Bomb Splash",					NONE, 0xA0, NO_LOOP },
	/* 0x2A */	{ "Mr. Blizzard Sink",					NONE, 0x00, NO_LOOP },
	/* 0x2B */	{ "Snowball hit Ground",				NONE, 0x00, NO_LOOP },
	/* 0x2C */	{ "Flying Bookend Launch",				NONE, 0x80, NO_LOOP },
	/* 0x2D */	{ "Mother Penguin Yell",				NONE, 0x00, NO_LOOP },
	/* 0x2E */	{ "Water Bomb Bounce",					NONE, 0x80, NO_LOOP },
	/* 0x2F */	{ "Enemy Detects Mario",				NONE, 0x00, NO_LOOP },
	/* 0x30 */	{ "Generic Enemy Stomped",				NONE, 0x80, NO_LOOP },
	/* 0x31 */	{ "Swoop Flap",							NONE, 0x00, NO_LOOP },
	/* 0x32 */	{ "Object Enters Water",				NONE, 0x40, NO_LOOP },
	/* 0x33 */	{ "Piranha Plant Shrink",				NONE, 0x40, NO_LOOP },
	/* 0x34 */	{ "Koopa the Quick Run",				NONE, 0x20, NO_LOOP },
	/* 0x35 */	{ "Koopa Walk",							NONE, 0x00, NO_LOOP },
	/* 0x36 */	{ "Bully Walk (2)",						NONE, 0x60, NO_LOOP },
	/* 0x37 */	{ "Dorrie",								NONE, 0x60, NO_LOOP },
	/* 0x38 */	{ "Bowser Laugh (Message)",				NONE, 0x80, NO_LOOP },
	/* 0x39 */	{ "Monkey Noises 2",					NONE, 0x00, NO_LOOP },
	/* 0x3A */	{ "Monkey Noises 1",					NONE, 0x00, NO_LOOP },
	/* 0x3B */	{ "Monkey Step",						NONE, 0x00, NO_LOOP },
	/* 0x3C */	{ "Monkey Step on Leaves",				NONE, 0x00, NO_LOOP },
	/* 0x3D */	{ "Koopa Talk",							NONE, 0xA0, NO_LOOP },
	/* 0x3E */	{ "Koopa Stomped",						NONE, 0xA0, NO_LOOP },
	/* 0x3F */	{ "Klepto Squawk (Low)",				NONE, 0x40, NO_LOOP },
	/* 0x40 */	{ "Klepto Squawk (High)",				NONE, 0x60, NO_LOOP },
	/* 0x41 */	{ "King Bob-Omb Talk",					NONE, 0x00, NO_LOOP },
	/* 0x42 */	{ "King Bob-Omb Hurt",					CONSTANT_VOLUME, 0x40, NO_LOOP },
	/* 0x43 */	{ "Scuttlebug Walk",					NONE, 0x40, NO_LOOP },
	/* 0x44 */	{ "Scuttlebug Detected Mario",			NONE, 0x40, NO_LOOP },
	/* 0x45 */	{ "Baby Penguin Yell",					NONE, 0x00, NO_LOOP },
	/* 0x46 */	{ "King Bob-Omb Jump Recover",			NONE, 0x80, NO_LOOP },
	/* 0x47 */	{ "King Whomp Defeated",				NONE, 0xC0, NO_LOOP },
	/* 0x48 */	{ "Boo Laugh (Long)",					NONE, 0x00, NO_LOOP },
	/* 0x49 */	{ "Swoop Shriek",						NONE, 0x00, NO_LOOP },
	/* 0x4A */	{ "Unagi Shriek",						NONE, 0x00, NO_LOOP },
	/* 0x4B */	{ "Eyerok Opening Eye",					NONE, 0x00, NO_LOOP },
	/* 0x4C */	{ "Mr. Blizzard Rise Up",				NONE, 0x00, NO_LOOP },
	/* 0x4D */	{ "Snufit Shoot",						NONE, 0x00, NO_LOOP },
	/* 0x4E */	{ "Skeeter Walk",						NONE, 0x00, NO_LOOP },
	/* 0x4F */	{ "Skeeter Glide",						NONE, 0x00, NO_LOOP },
	/* 0x50 */	UNDEFINED,
	/* 0x51 */	{ "Bird Chirp (2)",						NONE, 0x40, LOOP },
	/* 0x52 */	{ "Bird Chirp (3)",						NONE, 0x40, LOOP },
	/* 0x53 */	UNDEFINED,
	/* 0x54 */	{ "Piranha Plant Grow",					NONE, 0x20, NO_LOOP },
	/* 0x55 */	{ "Single Flame Spawn",					NONE, 0x80, NO_LOOP },
	/* 0x56 */	{ "Mad Piano Chomp",					VIBRATO, 0x40, NO_LOOP },
	/* 0x57 */	{ "Big Bully Pushed Back",				NONE, 0x00, NO_LOOP },
	/* 0x58 */	{ "Bob-Omb Talk",						NONE, 0x40, NO_LOOP },
	/* 0x59 */	{ "Spiny Bounce",						NONE, 0x10, NO_LOOP },
	/* 0x5A */	{ "Whomp Flop/Eyerok Hurt",				VIBRATO | CONSTANT_VOLUME, 0xC0, NO_LOOP },
	/* 0x5B */	{ "Eyerok Hand Destroyed",				VIBRATO, 0x00, NO_LOOP },
	/* 0x5C */	{ "Tiny Wiggler Walk",					NONE, 0x40, NO_LOOP },
	/* 0x5D */	{ "Heave-Ho Toss",						NONE, 0x40, NO_LOOP },
	/* 0x5E */	{ "Wiggler Defeated",					NONE, 0x80, NO_LOOP },
	/* 0x5F */	{ "Bowser Laugh (Intro)",				NONE, 0x80, NO_LOOP | FADE },
	/* 0x60 */	{ "Tiny Goomba Squished",				NONE, 0xB0, NO_LOOP },
	/* 0x61 */	{ "Giant Goomba Squished",				NONE, 0xB0, NO_LOOP },
	/* 0x62 */	{ "Swoop Defeated",						NONE, 0xB0, NO_LOOP },
	/* 0x63 */	{ "Koopa/Flyguy Defeated",				NONE, 0xB0, NO_LOOP },
	/* 0x64 */	{ "Snowman Bounce",						NONE, 0xC0, NO_LOOP },
	/* 0x65 */	{ "Snowman Head Exploded",				NONE, 0xD0, NO_LOOP },
	/* 0x66 */	{ "Bowser Teleport",					NONE, 0x80, NO_LOOP },
	/* 0x67 */	{ "Monty Mole Resurface",				NONE, 0x80, NO_LOOP },
	/* 0x68 */	{ "Eyerok Pound",						NONE, 0x40, NO_LOOP },
	/* 0x69 */	{ "King Whomp/Eyerok Grunt",			NONE, 0x40, NO_LOOP },
	/* 0x6A */	{ "MIPS Noises",						NONE, 0x00, NO_LOOP },
	/* 0x6B */	{ "Mr. I Spinning",						NONE, 0x00, NO_LOOP },
	/* 0x6C */	{ "MIPS Noises (in Water)",				NONE, 0x00, NO_LOOP },
	/* 0x6D */	{ "Eyerok Defeated",					NONE, 0x00, NO_LOOP },
	/* 0x6E */	{ "Chuckya Defeated",					CONSTANT_VOLUME, 0x00, NO_LOOP },
	/* 0x6F */	{ "Wiggler Talk",						NONE, 0x00, NO_LOOP },
	/* 0x70 */	{ "Wiggler Stomped",					NONE, 0x60, NO_LOOP },
	/* 0x71 */	{ "Wiggler Walk",						NONE, 0x20, NO_LOOP },
	/* 0x72 */	{ "Skeeter/Snufit Defeated",			NONE, 0xC0, NO_LOOP },
	/* 0x73 */	{ "Bubba Chomp",						NONE, 0x40, NO_LOOP },
	/* 0x74 */	{ "Enemy Defeated (Shrinking)",			NONE, 0x40, NO_LOOP }
};

static const SoundInfo bank6[] = {
	/* 0x00 */	{ "Bowser Fire Breath",					NONE, 0x00, LOOP },
	/* 0x01 */	{ "Unused Pulley/Elevator Sound",		NONE, 0x80, LOOP },
	/* 0x02 */	{ "Lakitu Siren",						NONE, 0x80, LOOP },
	/* 0x03 */	{ "Amp Buzzing",						NONE, 0x40, LOOP },
	/* 0x04 */	{ "Fire Spewer",						NONE, 0x80, LOOP },
	/* 0x05 */	{ "Rough Slide",						NONE, 0x00, LOOP },
	/* 0x06 */	{ "Heave-Ho Moving",					NONE, 0x40, LOOP },
	/* 0x07 */	{ "Unused Repeating Wheep Sound",		NONE, 0x80, LOOP },
	/* 0x08 */	{ "Bob-Omb Fuse",						NONE, 0x60, LOOP },
	/* 0x09 */	{ "Howling Wind",						NONE, 0x80, LOOP },
	/* 0x0A */	{ "Chuckya Moving",						NONE, 0x40, LOOP },
	/* 0x0B */	{ "Peach Floating Down",				NONE, 0x40, LOOP },
	/* 0x0C */	UNDEFINED,
	/* 0x0D */	UNDEFINED,
	/* 0x0E */	UNDEFINED,
	/* 0x0F */	UNDEFINED,
	/* 0x10 */	{ "Ambient Bird Sounds",				NONE, 0x40, LOOP }
};

static const SoundInfo bank7[] = {
	/* 0x00 */	{ "Menu Option Select",					NONE, 0xF8, NO_LOOP },
	/* 0x01 */	{ "Reversed Pause Sound",				NONE, 0xFF, NO_LOOP },
	/* 0x02 */	{ "Pause Sound",						NONE, 0xFF, NO_LOOP },
	/* 0x03 */	{ "Pause Sound (2)",					NONE, 0xFF, NO_LOOP },
	/* 0x04 */	{ "Message Appears",					NONE, 0x00, NO_LOOP },
	/* 0x05 */	{ "Message Dismissed",					NONE, 0x00, NO_LOOP },
	/* 0x06 */	{ "Camera Zoom In",						NONE, 0x00, NO_LOOP },
	/* 0x07 */	{ "Camera Zoom Out",					NONE, 0x00, NO_LOOP },
	/* 0x08 */	{ "Pinch Mario's Face",					NONE, 0x00, NO_LOOP },
	/* 0x09 */	{ "Let Go of Mario's Face",				NONE, 0x00, NO_LOOP },
	/* 0x0A */	{ "Mario Face - Cursor Appears",		NONE, 0x00, NO_LOOP },
	/* 0x0B */	{ "Mario Face - Cursor Hidden",			NONE, 0x00, NO_LOOP },
	/* 0x0C */	{ "Unknown Bling Sound",				NONE, 0x00, NO_LOOP },
	/* 0x0D */	{ "Health Get",							NONE, 0x00, NO_LOOP },
	/* 0x0E */	{ "Camera Buzzer",						NONE, 0x00, NO_LOOP },
	/* 0x0F */	{ "Camera Turn",						NONE, 0x00, NO_LOOP },
	/* 0x10 */	UNDEFINED,
	/* 0x11 */	{ "File Select Click",					NONE, 0x00, NO_LOOP },
	/* 0x12 */	{ "Unused Message Sound",				NONE, 0x00, NO_LOOP },
	/* 0x13 */	{ "Dialog Next Page",					NONE, 0x00, NO_LOOP },
	/* 0x14 */	{ "Coin + It's-a Me, Mario",			NONE, 0x00, NO_LOOP },
	/* 0x15 */	{ "Yoshi Gives 1UP",					NONE, 0x00, NO_LOOP },
	/* 0x16 */	{ "Enter Pipe",							NONE, 0xA0, NO_LOOP },
	/* 0x17 */	{ "Exit Pipe",							NONE, 0xA0, NO_LOOP },
	/* 0x18 */	{ "Bowser Laugh (Death Warp)",			NONE, 0x80, NO_LOOP },
	/* 0x19 */	{ "Enter Hole",							CONSTANT_VOLUME, 0x80, NO_LOOP },
	/* 0x1A */	{ "Change Camera",						NONE, 0x80, NO_LOOP },
	/* 0x1B */	{ "Unused Camera Sound 1",				NONE, 0x80, NO_LOOP },
	/* 0x1C */	{ "Unused Camera Sound 2",				NONE, 0x80, NO_LOOP },
	/* 0x1D */	{ "Pause Warp",							NONE, 0xB0, NO_LOOP },
	/* 0x1E */	{ "Star Spawn",							NONE, 0xFF, NO_LOOP },
	/* 0x1F */	{ "Thanks For Playing",					NONE, 0xFF, NO_LOOP },
	/* 0x20 */	{ "Read Sign (Single Blip)",			NONE, 0x80, NO_LOOP },
	/* 0x21 */	{ "Done Reading Sign (Single Blip)",	NONE, 0x80, NO_LOOP },
	/* 0x22 */	{ "Castle Warp (Got Wing Cap Star)",	NONE, 0x20, NO_LOOP },
	/* 0x23 */	{ "Okie Dokie",							NONE, 0xFF, NO_LOOP },
	/* 0x24 */	{ "Let's-a go!",						NONE, 0xFF, NO_LOOP },
	/* 0x30 */	UNDEFINED,
	/* 0x31 */	UNDEFINED,
	/* 0x32 */	UNDEFINED,
	/* 0x33 */	{ "Red Coin 1",							NONE, 0x90, NO_LOOP },
	/* 0x34 */	{ "Red Coin 2",							NONE, 0x90, NO_LOOP },
	/* 0x35 */	{ "Red Coin 3",							NONE, 0x90, NO_LOOP },
	/* 0x36 */	{ "Red Coin 4",							NONE, 0x90, NO_LOOP },
	/* 0x37 */	{ "Red Coin 5",							NONE, 0x90, NO_LOOP },
	/* 0x38 */	{ "Red Coin 6",							NONE, 0x90, NO_LOOP },
	/* 0x39 */	{ "Red Coin 7",							NONE, 0x90, NO_LOOP },
	/* 0x3A */	{ "Red Coin 8",							NONE, 0x90, NO_LOOP },
	/* 0x3B */	{ "Secret 1",							NONE, 0x20, NO_LOOP },
	/* 0x3C */	{ "Secret 2",							NONE, 0x20, NO_LOOP },
	/* 0x3D */	{ "Secret 3",							NONE, 0x20, NO_LOOP },
	/* 0x3E */	{ "Secret 4",							NONE, 0x20, NO_LOOP },
	/* 0x3F */	{ "Secret 5",							NONE, 0x20, NO_LOOP },
	/* 0x40 */	{ "Secret 6 (Unused)",					NONE, 0x20, NO_LOOP },
	/* 0x41 */	{ "Secret 7 (Unused)",					NONE, 0x20, NO_LOOP }
};

// Bank 8 is a copy of bank 3
// Bank 9 is a copy of bank 5

static const SoundInfo *soundBanks[] = {
	bank0, bank1, bank2, bank3, bank4, bank5, bank6, bank7, bank3, bank5
};

#define BANK_SIZE( n ) (sizeof( bank##n ) / sizeof( SoundInfo ))

static_assert( BANK_SIZE(0) == 0x67 + 1 );

static const size_t bankSizes[] = {
	BANK_SIZE(0), BANK_SIZE(1), BANK_SIZE(2), BANK_SIZE(3), BANK_SIZE(4), BANK_SIZE(5), BANK_SIZE(6), BANK_SIZE(7),
	BANK_SIZE(3), BANK_SIZE(5)
};

string getSoundName( uint soundId ) {
	if( soundId == 0 ) return "No Sound";

	const ubyte bank = soundId >> 28;
	const ubyte index = (soundId >> 16) & 0xFF;

	if( bank > 9 || index >= bankSizes[bank] || soundBanks[bank][index].name == nullptr ) {
		char buffer[7];
		std::sprintf( buffer, "0x%04x", soundId >> 16 );
		buffer[6] = '\0';
		return string( buffer );
	}

	return string( soundBanks[bank][index].name );
}

static constexpr size_t totalSize = sizeof( bank0 ) + sizeof( bank1 ) + sizeof( bank2 ) + sizeof( bank3 ) + sizeof( bank4 ) + sizeof( bank5 ) + sizeof( bank6 ) + sizeof( bank7 );

std::vector<std::pair<uint,const char*>> getAllNamedSounds() {
	std::vector<std::pair<uint,const char*>> namedSounds;
	namedSounds.reserve( totalSize / 4 );
	for( uint i = 0; i < 8; i++ ) {
		for( uint j = 0; j < bankSizes[i]; j++ ) {
			const SoundInfo &sound = soundBanks[i][j];
			if( sound.name == nullptr ) continue;
			namedSounds.push_back({(
				(i << 28) |
				((uint)sound.upperFlags << 24) |
				((uint)j << 16) |
				((uint)sound.priority << 8) |
				((uint)sound.lowerFlags << 4) |
				0x1
			), sound.name });
		}
	}
	return namedSounds;
}
